VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "PathLinks"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit
'=============
'PathLinks 1.2
'=============
'
'Yes, you've heard rumors.  Well here it is: The Amazing PathLinks!
'
'Why PathLinks?
'==============
'
'People have been muddling along for years putting data files into App.Path
'and getting away with it fine much of the time.  But along came Windows
'Vista and the rules seemed to suddenly change!
'
'So now you can't safely put your read/write files into folders under
'App.Path anymore.  What to do?
'
'
'PathLinks To The Rescue!
'========================
'
'At program startup you make a call to PathLinks.Publish(), passing it the
'relative path of a folder or folders you install under App.Path when you
'install your program.
'
'The first run after installation should be elevated.  After that all users
'can reach your data subfolders without elevation.
'
'   * If running on a pre-Vista OS:
'
'       + The specified data subfolders are created under App.Path if they
'         do not already exist.
'
'       + A good result (plPublishEnum.plpContinue) is returned.
'
'   * If running on Vista or later for the first time:
'
'       + If running non-elevated an error result
'         (plPublishEnum.plpElevationRequired) is returned and you should ask
'         the user to run once as administrator.
'
'       + If running first time and elevated then PathLinks.Publish() will
'         try to:
'
'           o Locate the "Public" special folder.
'
'               NOTE: If this fails it stops and returns an error
'                     (plPublishEnum.plpFailed) which is fatal but should
'                     be extremely rare!
'
'           o Create a folder under "Public" named:
'
'                   - <App.CompanyName>\<App.ProductName>
'                   - <App.CompanyName>\<App.ExeName>
'               or:
'                   - <App.ExeName>
'
'             Depending on whether you have CompanyName and ProductName defined
'             in your project.
'
'           o Go through the list of one or more App.Path data subfolders as
'             passed to PathLinks.Publish() and for each one:
'
'                   - Move that subfolder tree from App.Path to the new
'                     folder in "Public."
'
'                   - Create an empty subfolder under the new folder in
'                     "Public" if no existing data folder of that name was
'                     found under App.Path.
'
'                   - Create a symlink in App.Path named for the top-level
'                     name in the data subfolder path that points to the new
'                     location under the new folder in "Public."
'
'           o Return a good result (plPublishEnum.plpContinue).
'
'   * If running on Vista or later a subsequent time either elevated or not,
'     a good result (plPublishEnum.plpContinue) is returned.
'
'
'Misc. Notes
'===========
'
'   * Do not "publish" subfolders under App.Path that contain DLLs, OCXs, etc.
'     This is just for data and data folders you need read/write access to for
'     any user running your program.
'
'   * Data files placed "naked" right under App.Path without a subfolder can
'     not be helped by PathLinks.  If necessary rewrite your program to use
'     these from inside a subfolder.
'
'   * Create your setup package as you would on Windows 2000 or XP.  Include
'     any data files in the package to be deployed to a subfolder under
'     App.Path as always.
'
'   * The only extra step after installing is a user must run your program
'     elevated one time before anyone uses it normally.
'
'   * Once again, pre-Vista PathLinks does nothing except create empty data
'     subfolders for you as required.  From there your program runs as
'     always.
'
'   * You could also write a separate small program that just calls
'     PathLinks.Publish() to be run elevated right after installing.  Then
'     your actual application program or programs will not need PathLinks in
'     them at all!
'
'
'Passing The Folder List
'=======================
'
'PathLinks.Publish() expects a Variant parameter.  This Variant should contain
'either a single String value or else an array of String values:
'
'   result = PathLinks.Publish("temp")
'
'   result = PathLinks.Publish(Array("temp", "dictionary\db", "images"))
'
'
'What Does This Do For You?
'==========================
'
'If running pre-Vista, nothing.
'
'If running on Vista or later:
'
'   PathLinks will move your data subfolders into a program work folder that
'   it creates under the computer's "Public" folder.  This is a read/write
'   location for all logged-on users without elevation.
'
'   Then PathLinks will create a symlink for each subfolder in App.Path that
'   points to the new location of your data subfolder.
'
'   Your program can use App.Path as always, and should not notice any
'   difference!
'

'---- Private Declarations -------------------------------------------

Private Const S_OK As Long = 0
Private Const SYMBOLIC_LINK_FLAG_DIRECTORY As Long = 1
Private Const FOLDERID_Public As String = "{DFDF76A2-C82A-4D63-906A-5644AC457385}"

Private Declare Function CLSIDFromString Lib "ole32" ( _
    ByVal lpszGuid As Long, _
    ByVal pGuid As Long) As Long

Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" ( _
    ByVal Destination As Long, _
    ByVal Source As Long, _
    ByVal Length As Long)

Private Declare Sub CoTaskMemFree Lib "ole32" (ByVal hMem As Long)

Private Declare Function CreateSymbolicLink Lib "kernel32" Alias "CreateSymbolicLinkW" ( _
    ByVal lpSymlinkFileName As Long, _
    ByVal lpTargetFileName As Long, _
    ByVal dwFlags As Long) As Long

Private Declare Function GetVersion Lib "kernel32" () As Long

Private Declare Function IsUserAnAdmin Lib "shell32" () As Long

Private Declare Function lstrlen Lib "kernel32" Alias "lstrlenW" (ByVal ptr As Long) As Long

Private Declare Function SHGetKnownFolderPath Lib "shell32" ( _
    ByVal rfid As Long, _
    ByVal dwFlags As Long, _
    ByVal hToken As Long, _
    ByRef ppszPath As Long) As Long

Private Type GUID
    Data1 As Long
    Data2 As Integer
    Data3 As Integer
    Data4(7) As Byte
End Type

Private Enum plStatusEnum
    plsOldOS
    plsWin6PlusElevated
    plsWin6PlusElevationRequired
    plsFailed
End Enum

'---- Public Enums ---------------------------------------------------

Public Enum plPublishEnum
    plpContinue
    plpElevationRequired
    plpFailed
End Enum

#If False Then 'IDE trick to preserve case of Enum values.
    Dim plpContinue, plpElevationRequired, plpFailed
#End If

'---- Private Data ---------------------------------------------------

Private mPfStatus As plStatusEnum
Private mStrPublicFolder As String

'---- Public Properties and Methods ----------------------------------

Public Function Publish(ByVal SubFolders As Variant) As plPublishEnum
    Dim I As Integer
    Dim strCDCache As String
    
    If (VarType(SubFolders) And vbArray) = vbArray Then
        For I = LBound(SubFolders) To UBound(SubFolders)
            If VarType(SubFolders(I)) <> vbString Then
                GoTo BadArgument
            End If
        Next
    Else
        If VarType(SubFolders) <> vbString Then
            GoTo BadArgument
        End If
    End If
    
    Select Case mPfStatus
        Case plsOldOS
            If (VarType(SubFolders) And vbArray) = vbArray Then
                For I = LBound(SubFolders) To UBound(SubFolders)
                    CreateSubFolder App.Path, SubFolders(I)
                Next
            Else
                CreateSubFolder App.Path, SubFolders
            End If
            
            Publish = plpContinue
        
        Case plsWin6PlusElevationRequired
            strCDCache = CurDir$()
            ChDir App.Path
            ChDrive App.Path
            Publish = plpElevationRequired
            
            If (VarType(SubFolders) And vbArray) = vbArray Then
                For I = LBound(SubFolders) To UBound(SubFolders)
                    If Not Exists(SubFolders(I)) Then Exit Function
                Next
            Else
                If Not Exists(SubFolders) Then Exit Function
            End If
            
            Publish = plpContinue
            ChDir strCDCache
            ChDrive strCDCache
        
        Case plsWin6PlusElevated
            If (VarType(SubFolders) And vbArray) = vbArray Then
                For I = LBound(SubFolders) To UBound(SubFolders)
                    CreateMoveLink SubFolders(I)
                Next
            Else
                CreateMoveLink SubFolders
            End If
            
            Publish = plpContinue
        
        Case plsFailed
            Publish = plpFailed
    End Select
    
    Exit Function
    
BadArgument:
    Err.Raise 5, _
              "PathLinks.Publish", _
              "String or Variant Array of Strings required."
End Function

'---- Private Methods ------------------------------------------------

Private Sub CreateMoveLink(ByVal SubFolder As String)
    If Left$(SubFolder, 1) = "\" Then SubFolder = Mid$(SubFolder, 2)
    CreateSubFolder App.Path, SubFolder
    CreateSubFolder mStrPublicFolder, SubFolder
    If Exists(App.Path & "\" & SubFolder) Then MoveFiles SubFolder
    CreateSymbolicLink StrPtr(App.Path & "\" & SubFolder), _
                       StrPtr(mStrPublicFolder & "\" & SubFolder), _
                       SYMBOLIC_LINK_FLAG_DIRECTORY
End Sub

Private Sub CreateSubFolder(ByVal ParentPath As String, ByVal SubFolder As String)
    Dim strBuiltPath As String
    Dim strNodes() As String
    Dim lngNode As Long
    
    strNodes = PathNodes(SubFolder)
    strBuiltPath = ParentPath
    For lngNode = 0 To UBound(strNodes)
        strBuiltPath = strBuiltPath & "\" & strNodes(lngNode)
        If Not Exists(strBuiltPath) Then
            MkDir strBuiltPath
        End If
    Next
End Sub

Private Function Exists(ByVal FileOrFolder) As Boolean
    On Error Resume Next
    GetAttr FileOrFolder
    Exists = Err.Number = 0
End Function

Private Function GetKnownFolder(ByVal KnownFolderID As String) As String
    Dim guidFOLDERID As GUID
    Dim pPath As Long
    Dim lngLen As Long

    If CLSIDFromString(StrPtr(KnownFolderID), VarPtr(guidFOLDERID)) = 0 Then
        If SHGetKnownFolderPath(VarPtr(guidFOLDERID), 0, 0, pPath) = 0 Then
            lngLen = lstrlen(pPath)
            GetKnownFolder = String$(lngLen, 0)
            CopyMemory StrPtr(GetKnownFolder), pPath, 2 * lngLen
            CoTaskMemFree pPath
            Exit Function
        End If
    End If
    
    GetKnownFolder = vbNullString
End Function

Private Sub MoveFiles(ByVal Folder As String)
    Dim colSubFolders As Collection
    Dim strName As String
    Dim strFQName As String
    Dim varSubFolder As Variant
    
    Set colSubFolders = New Collection
    strName = Dir$(App.Path & "\" & Folder & "\*", vbNormal Or vbDirectory)
    
    Do Until Len(strName) = 0
        strFQName = App.Path & "\" & Folder & "\" & strName
        If (GetAttr(strFQName) And vbDirectory) = vbDirectory Then
            If Not (strName = "." Or strName = "..") Then
                colSubFolders.Add Folder & "\" & strName
            End If
        Else
            FileCopy strFQName, mStrPublicFolder & "\" & Folder & "\" & strName
            Kill strFQName
        End If
        strName = Dir$()
    Loop
        
    For Each varSubFolder In colSubFolders
        MkDir mStrPublicFolder & "\" & varSubFolder
        MoveFiles varSubFolder
    Next
    
    RmDir App.Path & "\" & Folder
End Sub

Private Function PathNodes(ByVal Path As String) As String()
    Dim strNodes() As String
    Dim lngNodeLast As Long
    
    strNodes = Split(Path, "\")
    lngNodeLast = UBound(strNodes)
    If lngNodeLast > -1 Then
        'Trim trailing empty node if any.
        If Len(strNodes(lngNodeLast)) = 0 Then
            ReDim Preserve strNodes(lngNodeLast - 1)
        End If
    End If
    PathNodes = strNodes
End Function

'---- Event Handlers -------------------------------------------------
    
Private Sub Class_Initialize()
    Dim strPublicBase As String
    Dim pPath As Long
    
    If (GetVersion() And &HFF&) >= 6 Then
        If IsUserAnAdmin() Then
            mStrPublicFolder = GetKnownFolder(FOLDERID_Public)
            If StrPtr(mStrPublicFolder) = 0 Then
                mPfStatus = plsFailed
            Else
                mPfStatus = plsWin6PlusElevated
                If Len(App.CompanyName) > 0 Then
                    strPublicBase = App.CompanyName & "\"
                End If
                If Len(App.ProductName) > 0 Then
                    strPublicBase = strPublicBase & App.ProductName
                Else
                    strPublicBase = strPublicBase & App.EXEName
                End If
                CreateSubFolder mStrPublicFolder, strPublicBase
                mStrPublicFolder = mStrPublicFolder & "\" & strPublicBase
            End If
        Else
            mPfStatus = plsWin6PlusElevationRequired
        End If
    Else
        mPfStatus = plsOldOS
    End If
End Sub
